// Replacement firmware for Chinese FM transmitter
// Version 0.1
// Scan: Press Up+Down
// For ATMega48 8MHz internal clock, CKDIV=0
// CC Furrtek 07/2012

// Digit segments, sink outputs
#define SEG_A	PD7
#define SEG_B	PB2
#define SEG_C	PD6
#define SEG_D	PB1
#define SEG_E	PD5
#define SEG_F	PD3
#define SEG_G	PB4
#define SEG_DOT	PB5

// Digit selection, source outputs
#define	D_SEL0	PB6	// Leftmost digit ("1")
#define	D_SEL1	PB3
#define	D_SEL2	PB7
#define	D_SEL3	PB0 // Rightmost digit

#define BEEP	PC0	// Scan tone output pin

#define BTN_UP	PC1	// UP button, active low
#define BTN_DN	PC2	// DOWN button, active low

// BH1415 communication pins
#define	CE		PC3
#define DATA	PC4
#define CLK		PC5

// Digit segment definitions (invariable)
#define DIG_0	0xFF ^ (_BV(SEG_A) | _BV(SEG_B) | _BV(SEG_C) | _BV(SEG_D) | _BV(SEG_E) | _BV(SEG_F))
#define DIG_1	0xFF ^ (_BV(SEG_B) | _BV(SEG_C))
#define DIG_2	0xFF ^ (_BV(SEG_A) | _BV(SEG_B) | _BV(SEG_D) | _BV(SEG_E) | _BV(SEG_G))
#define DIG_3	0xFF ^ (_BV(SEG_A) | _BV(SEG_B) | _BV(SEG_C) | _BV(SEG_D) | _BV(SEG_G))
#define DIG_4	0xFF ^ (_BV(SEG_B) | _BV(SEG_C) | _BV(SEG_F) | _BV(SEG_G))
#define DIG_5	0xFF ^ (_BV(SEG_A) | _BV(SEG_C) | _BV(SEG_D) | _BV(SEG_F) | _BV(SEG_G))
#define DIG_6	0xFF ^ (_BV(SEG_A) | _BV(SEG_C) | _BV(SEG_D) | _BV(SEG_E) | _BV(SEG_F) | _BV(SEG_G))
#define DIG_7	0xFF ^ (_BV(SEG_A) | _BV(SEG_B) | _BV(SEG_C))
#define DIG_8	0xFF ^ (_BV(SEG_A) | _BV(SEG_B) | _BV(SEG_C) | _BV(SEG_D) | _BV(SEG_E) | _BV(SEG_F) | _BV(SEG_G))
#define DIG_9	0xFF ^ (_BV(SEG_A) | _BV(SEG_B) | _BV(SEG_C) | _BV(SEG_D) | _BV(SEG_F) | _BV(SEG_G))

// "Splash digits" (FTK)
#define DIG_F	0xFF ^ (_BV(SEG_A) | _BV(SEG_E) | _BV(SEG_F) | _BV(SEG_G))
#define DIG_T	0xFF ^ (_BV(SEG_A) | _BV(SEG_E) | _BV(SEG_F))
#define DIG_K	0xFF ^ (_BV(SEG_B) | _BV(SEG_C) | _BV(SEG_E) | _BV(SEG_F) | _BV(SEG_G))

#define F_CPU 8000000
#define FREQ_MAX 1200	// Max frequency 120MHz
#define FREQ_MIN 700	// Min frequency 70MHz

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <inttypes.h>
#include <avr/eeprom.h>

uint16_t EEMEM lastfreq;	// EEPROM word for last frequency used
uint8_t volatile digit;		// Number of the currently displayed digit
uint16_t volatile val;		// Frequency (in 100kHz steps)
uint16_t volatile cmptv;	// Splash display timeout counter
uint8_t volatile cmpt=0;	// Scan speed timer
uint8_t volatile cmptb=0;	// Scan tone timer
uint8_t volatile scan=0;	// Scanning on/off (boolean)

const uint8_t nums[14] = {	DIG_0,
							DIG_1,
							DIG_2,
							DIG_3,
							DIG_4,
							DIG_5,
							DIG_6,
							DIG_7,
							DIG_8,
							DIG_9,
							0xFF,	// Used to turn of the "1" digit
							DIG_F,
							DIG_T,
							DIG_K,
							};

// Bit-bangs 16bit word to BH1415 (can be sped up to 2us)
void sendBH1415(uint16_t sval) {
	uint8_t bits;

	PORTC |= _BV(CE);		// CE high
	_delay_us(10);
	for (bits=0;bits<16;bits++) {
		if ((sval>>bits) & 1) {	// Shift out data LSB first
			PORTC |= _BV(DATA);
		} else {
			PORTC &= ~_BV(DATA);
		}
		_delay_us(10);
		PORTC |= _BV(CLK);	// Clock high
		_delay_us(10);
		PORTC &= ~_BV(CLK);	// Clock low
	}
	_delay_us(10);
	PORTC &= ~_BV(CE);		// CE low
}

// Display and scanning
ISR(TIMER0_OVF_vect) {
	uint8_t PORTBv,PORTDv;	// Used to make values for PORTB and PORTD
	uint8_t vald;			// Value of current digit

	PORTBv = 0;
	PORTDv = 0;
	if (digit == 0) {				// "1" digit (hundreds)
		if ((val > 999) && (!cmptv)) {
			vald = 1;				// Display "1" is freq > 99.9MHz
		} else {
			vald = 10;				// Or turn off
		}
		PORTBv |= _BV(D_SEL0);
	} else if (digit == 1) {
		if (cmptv) {
			vald = 11;				// "F" if displaying splash
		} else {
			vald = (val/100)%10;	// Or display tens
		}
		PORTBv |= _BV(D_SEL1);
	} else if (digit == 2) {
		if (cmptv) {
			vald = 12;				// "T" if displaying splash
		} else {
			vald = (val/10)%10;		// Or display units
		}
		PORTBv |= _BV(D_SEL2);
	} else {
		if (cmptv) {
			vald = 13;				// "K" if displaying splash
		} else {
			vald = val%10;			// Or display 0.1MHz's
		}
		PORTBv |= _BV(D_SEL3);
	}

	if (cmptv) {
		cmptv--;				// Splash timing
		PORTBv |= _BV(PB5);		// Turn off dot
	}

	PORTBv |= nums[vald] & 0b00010110;	// Mask bits for PORTB
	PORTDv |= nums[vald] & 0b11101000;	// Mask bits for PORTD
	PORTB = PORTBv;
	PORTD = PORTDv;

	sendBH1415(val | 0b0100100000000000);	// Stereo, normal PD op and test bits

	digit++;					// Next digit
	if (digit == 4) digit = 0;	// Loop to first one if last one

	if (scan) {					// Scanning...
		if (cmpt == 100) {		// Add 0.1MHz to frequency after timing
			cmpt = 0;
			val++;
		}
		cmpt++;
		if (val == FREQ_MAX) {	// Stop scanning if at end of freq range
			scan = 0;
			DDRC = 0b00111000;	// Tone out as input
		}
		cmptb++;
		if (cmptb & 0x80) PORTC ^= _BV(BEEP);	// Output tone bursts
	}
}

int main(void) {
	WDTCSR = (1<<WDCE) | (1<<WDE);	// Put watchdog asleep (not good...)
	WDTCSR = 0x00;

	DDRB = 0xFF;	// Port B and D are all outputs
	DDRD = 0xFF;
	DDRC = _BV(CE) | _BV(CLK) | _BV(DATA);
	PORTC = _BV(BTN_UP) | _BV(BTN_DN);		// Button pull-ups

	val = eeprom_read_word(&lastfreq);		// Read last used freq from EEPROM
	if ((val < FREQ_MIN) || (val > FREQ_MAX)) val = FREQ_MIN;	// Sanitation

	TCCR0A = 0b00000000;	// Normal Timer0 operation
	TCCR0B = 0b00000010;	// Prescaler
	TIMSK0 = 0b00000001;	// Overflow interrupt please

	cmptv = 4000;			// Splash display timeout

	_delay_ms(100);			// Wait for settling

	sei();

	// Main loop (inputs checking)
	for (;;) {
		if (!bit_is_set(PINC,BTN_UP)) {
			scan = 0;		// Stop scan
			val++;			// Add 0.1MHz
			if (val >= FREQ_MAX) val = FREQ_MIN;	// Loop back
			_delay_us(50);	// Cheap debounce
			// Save freq value if button released
			if (bit_is_set(PINC,BTN_UP)) eeprom_write_word(&lastfreq,val);
		}
		if (!bit_is_set(PINC,BTN_DN)) {
			scan = 0;		// Stop scan
			if (!bit_is_set(PINC,BTN_UP)) {
				scan = 1;	// UP button pressed too, start scan
				DDRC = _BV(CE) | _BV(CLK) | _BV(DATA) | _BV(BEEP);	// Set tone output
				val = FREQ_MIN;	// Start at lowest frequency
				while (!bit_is_set(PINC,BTN_DN)) {};	// Wait for button release
				_delay_us(50);	// Cheap debounce
			} else {
				val--;	// Substract 0.1MHz
				if (val <= FREQ_MIN) val = FREQ_MAX;	// Loop back
			}
			_delay_us(50);	// Cheap debounce
			// Save freq value if button released
			if (bit_is_set(PINC,BTN_DN)) eeprom_write_word(&lastfreq,val);
		}
	}
    return 0;
}
